@@ -45,15 +45,13 @@ RfConfig *rf_config_new(const char *config_path)
 
 char *rf_config_get_card_path(RfConfig *this)
 {
-	char *def = g_strdup("/dev/dri/card0");
-
-	g_return_val_if_fail(RF_IS_CONFIG(this), def);
+	g_return_val_if_fail(RF_IS_CONFIG(this), NULL);
 
 	g_autoptr(GError) error = NULL;
 	char *card =
 		g_key_file_get_string(this->f, RF_CONFIG_GROUP, "card", &error);
-	if (error != NULL)
-		return def;
+	if (error != NULL || card == NULL || card[0] == '\0')
+		return NULL;
 	char *card_path = g_strdup_printf("/dev/dri/%s", card);
 	g_free(card);
 	return card_path;
@@ -61,30 +59,35 @@ char *rf_config_get_card_path(RfConfig *this)
 
 char *rf_config_get_connector(RfConfig *this)
 {
-	char *def = g_strdup("eDP-1");
-
-	g_return_val_if_fail(RF_IS_CONFIG(this), def);
+	g_return_val_if_fail(RF_IS_CONFIG(this), NULL);
 
 	g_autoptr(GError) error = NULL;
 	char *connector = g_key_file_get_string(
 		this->f, RF_CONFIG_GROUP, "connector", &error
 	);
-	if (error != NULL)
-		return def;
+	if (error != NULL || connector == NULL || connector[0] == '\0')
+		return NULL;
 	return connector;
 }
 
-int rf_config_get_device_id(RfConfig *this)
+unsigned int rf_config_get_rotation(RfConfig *this)
 {
-	g_return_val_if_fail(RF_IS_CONFIG(this), -1);
+	g_return_val_if_fail(RF_IS_CONFIG(this), 0);
 
 	g_autoptr(GError) error = NULL;
-	int device_id = g_key_file_get_integer(
-		this->f, RF_CONFIG_GROUP, "device-id", &error
+	unsigned int rotation = g_key_file_get_integer(
+		this->f, RF_CONFIG_GROUP, "rotation", &error
 	);
 	if (error != NULL)
-		return -1;
-	return device_id;
+		return 0;
+	if (rotation % 90 != 0) {
+		g_warning(
+			"Got invalid monitor rotation angle %u, valid angles are clockwise 0, 90, 180, 270.",
+			rotation
+		);
+		rotation = rotation / 90 * 90;
+	}
+	return rotation % 360;
 }
 
 unsigned int rf_config_get_desktop_width(RfConfig *this)
@@ -139,26 +142,6 @@ int rf_config_get_monitor_y(RfConfig *this)
 	return monitor_y;
 }
 
-unsigned int rf_config_get_rotation(RfConfig *this)
-{
-	g_return_val_if_fail(RF_IS_CONFIG(this), 0);
-
-	g_autoptr(GError) error = NULL;
-	unsigned int rotation = g_key_file_get_integer(
-		this->f, RF_CONFIG_GROUP, "rotation", &error
-	);
-	if (error != NULL)
-		return 0;
-	if (rotation % 90 != 0) {
-		g_warning(
-			"Got invalid monitor rotation angle %u, valid angles are clockwise 0, 90, 180, 270.",
-			rotation
-		);
-		rotation = rotation / 90 * 90;
-	}
-	return rotation % 360;
-}
-
 unsigned int rf_config_get_default_width(RfConfig *this)
 {
 	g_return_val_if_fail(RF_IS_CONFIG(this), 0);
@@ -185,6 +168,19 @@ unsigned int rf_config_get_default_height(RfConfig *this)
 	return default_height;
 }
 
+bool rf_config_get_cursor(RfConfig *this)
+{
+	g_return_val_if_fail(RF_IS_CONFIG(this), true);
+
+	g_autoptr(GError) error = NULL;
+	gboolean cursor = g_key_file_get_boolean(
+		this->f, RF_CONFIG_GROUP, "cursor", &error
+	);
+	if (error != NULL)
+		return true;
+	return cursor ? true : false;
+}
+
 unsigned int rf_config_get_fps(RfConfig *this)
 {
 	g_return_val_if_fail(RF_IS_CONFIG(this), 30);
@@ -218,7 +214,7 @@ char *rf_config_get_password(RfConfig *this)
 	char *password = g_key_file_get_string(
 		this->f, RF_CONFIG_GROUP, "password", &error
 	);
-	if (error != NULL)
+	if (error != NULL || password == NULL || password[0] == '\0')
 		return NULL;
 	return password;
 }
