@@ -1,5 +1,6 @@
 #include <stdint.h>
 #include <stdbool.h>
+#include <gio/gio.h>
 #include <gio/gunixfdmessage.h>
 #include <linux/uinput.h>
 
@@ -7,6 +8,9 @@
 #include "rf-buffer.h"
 #include "rf-streamer.h"
 
+#define KEYBOARD_MAX_EVENTS 2
+#define POINTER_MAX_EVENTS 7
+
 struct _RfStreamer {
 	GObject parent_instance;
 	RfConfig *config;
@@ -24,99 +28,60 @@ struct _RfStreamer {
 	int monitor_y;
 	unsigned int rotation;
 	// These are the real size of monitor and have nothing with VNC.
-	unsigned int width;
-	unsigned int height;
+	uint32_t frame_width;
+	uint32_t frame_height;
 	bool running;
 };
 G_DEFINE_TYPE(RfStreamer, rf_streamer, G_TYPE_OBJECT)
 
-enum { SIG_START, SIG_STOP, SIG_FRAME, N_SIGS };
+enum { SIG_START, SIG_STOP, SIG_FRAME, SIG_CARD_PATH, N_SIGS };
 
 static unsigned int sigs[N_SIGS] = { 0 };
 
-static void _send_input_request(
-	RfStreamer *this,
-	struct input_event *ies,
-	const size_t length
-)
+static void
+_send_input_msg(RfStreamer *this, struct input_event *ies, const size_t length)
 {
-	char request = RF_REQUEST_TYPE_INPUT;
+	ssize_t ret = 0;
 	GOutputStream *os =
 		g_io_stream_get_output_stream(G_IO_STREAM(this->connection));
-	ssize_t ret = 0;
-
-	ret = g_output_stream_write(os, &request, sizeof(request), NULL, NULL);
-	if (ret == 0) {
-		goto disconnected;
-	} else if (ret < 0) {
-		g_warning("Input: Failed to send input request to socket.");
-		goto stop;
-	}
-
-	ret = g_output_stream_write(os, &length, sizeof(length), NULL, NULL);
-	if (ret == 0) {
-		goto disconnected;
-	} else if (ret < 0) {
-		g_warning(
-			"Input: Failed to send input events length to socket."
-		);
-		goto stop;
-	}
 
+	ret = rf_send_header(this->connection, RF_MSG_TYPE_INPUT, length);
+	if (ret <= 0)
+		goto out;
 	ret = g_output_stream_write(os, ies, length * sizeof(*ies), NULL, NULL);
-	if (ret == 0) {
-		goto disconnected;
-	} else if (ret < 0) {
-		g_warning(
-			"Input: Failed to send %lu * %ld bytes input events to socket.",
+
+out:
+	if (ret <= 0) {
+		g_warning("Input: Failed to send input events: %ld.", ret);
+		if (ret == 0)
+			g_warning("ReFrame Streamer disconnected.");
+		rf_streamer_stop(this);
+	} else {
+		g_debug("Input: Sent %ld * %ld bytes input event.",
 			length,
-			sizeof(*ies)
-		);
-		goto stop;
+			sizeof(*ies));
 	}
-
-	g_debug("Input: Sent %ld * %ld bytes input request.",
-		length,
-		sizeof(*ies));
-
-	return;
-
-disconnected:
-	g_warning("ReFrame Streamer disconnected.");
-stop:
-	rf_streamer_stop(this);
 }
 
-static gboolean _send_frame_request(gpointer data)
+static gboolean _send_frame_msg(gpointer data)
 {
 	RfStreamer *this = data;
-	char request = RF_REQUEST_TYPE_FRAME;
-	GOutputStream *os =
-		g_io_stream_get_output_stream(G_IO_STREAM(this->connection));
 	ssize_t ret = 0;
-
-	ret = g_output_stream_write(os, &request, sizeof(request), NULL, NULL);
-	if (ret == 0) {
-		g_warning("ReFrame Streamer disconnected.");
-		goto stop;
-	} else if (ret < 0) {
-		g_warning("Frame: Failed to send frame request to socket.");
-		goto stop;
+	ret = rf_send_header(this->connection, RF_MSG_TYPE_FRAME, 0);
+	if (ret <= 0) {
+		g_warning("Frame: Failed to send frame message: %ld.", ret);
+		if (ret == 0)
+			g_warning("ReFrame Streamer disconnected.");
+		rf_streamer_stop(this);
+	} else {
+		g_debug("Frame: Sent frame message.");
+		this->last_frame_time = g_get_monotonic_time();
+		this->timer_id = 0;
 	}
-
-	g_debug("Frame: Sent frame request.");
-
-	this->last_frame_time = g_get_monotonic_time();
-	this->timer_id = 0;
-
-	return G_SOURCE_REMOVE;
-
-stop:
-	rf_streamer_stop(this);
 	return G_SOURCE_REMOVE;
 }
 
-static void _schedule_frame_request(RfStreamer *this)
+static void _schedule_frame_msg(RfStreamer *this)
 {
 	if (this->timer_id != 0)
 		return;
@@ -126,16 +91,160 @@ static void _schedule_frame_request(RfStreamer *this)
 	if (delta < this->max_interval) {
 		this->timer_id = g_timeout_add(
 			(this->max_interval - delta) / 1000,
-			_send_frame_request,
+			_send_frame_msg,
 			this
 		);
 	} else {
 		if (this->last_frame_time != -1)
 			g_warning("Frame: Converting frame too slow.");
-		this->timer_id = g_timeout_add(1, _send_frame_request, this);
+		this->timer_id = g_timeout_add(1, _send_frame_msg, this);
 	}
 }
 
+static ssize_t _on_buffer(GSocketConnection *connection, RfBuffer *b)
+{
+	ssize_t ret = 0;
+	GSocket *socket = g_socket_connection_get_socket(connection);
+	GInputVector iov = { &b->md, sizeof(b->md) };
+	GSocketControlMessage **msgs = NULL;
+	int n_msgs = 0;
+
+	b->md.length = 0;
+	for (int i = 0; i < RF_MAX_FDS; ++i)
+		b->fds[i] = -1;
+
+	ret = g_socket_receive_message(
+		socket, NULL, &iov, 1, &msgs, &n_msgs, NULL, NULL, NULL
+	);
+	if (ret <= 0)
+		return ret;
+
+	// We should only receive 1 message each time.
+	if (n_msgs != 1) {
+		g_warning("Frame: Expect 1 fd message but got %d.", n_msgs);
+		return -2;
+	}
+
+	if (!G_IS_UNIX_FD_MESSAGE(msgs[n_msgs - 1])) {
+		g_warning("Frame: Failed to get fd message.");
+		return -2;
+	}
+
+	// We don't need to free GUnixFDList because
+	// GUnixFDMessage does not return a reference so we
+	// are not taking ownership of it.
+	//
+	// See <https://docs.gtk.org/gio-unix/type_func.FDMessage.get_fd_list.html>.
+	GUnixFDMessage *msg = G_UNIX_FD_MESSAGE(msgs[n_msgs - 1]);
+	GUnixFDList *fds = g_unix_fd_message_get_fd_list(msg);
+	for (int i = 0; i < b->md.length; ++i) {
+		b->fds[i] = g_unix_fd_list_get(fds, i, NULL);
+		// Some error happens.
+		if (b->fds[i] == -1) {
+			g_warning(
+				"Frame: Expect %d fds but got %d.",
+				b->md.length,
+				i
+			);
+			b->md.length = i;
+			break;
+		}
+	}
+
+	for (int i = 0; i < n_msgs; ++i)
+		g_object_unref(msgs[i]);
+	g_free(msgs);
+
+	rf_buffer_debug(b);
+	return ret;
+}
+
+static ssize_t _on_frame_msg(RfStreamer *this)
+{
+	g_debug("Frame: Received frame message.");
+
+	size_t length = 0;
+	ssize_t ret = 0;
+	GInputStream *is =
+		g_io_stream_get_input_stream(G_IO_STREAM(this->connection));
+
+	ret = g_input_stream_read(is, &length, sizeof(length), NULL, NULL);
+	if (ret <= 0)
+		goto out;
+	if (length <= 0 || length >= 3) {
+		g_warning("Frame: Got invalid buffers length %ld.", length);
+		goto out;
+	}
+
+	RfBuffer bufs[RF_MAX_BUFS];
+	for (size_t i = 0; i < length; ++i) {
+		ret = _on_buffer(this->connection, &bufs[i]);
+		if (ret <= 0)
+			goto out;
+	}
+
+	RfBuffer *primary = &bufs[0];
+	// Currently we assume primary size is always the same as monitor
+	// size. However, maybe we should use CRTC size actually?
+	uint32_t frame_width = primary->md.width;
+	uint32_t frame_height = primary->md.height;
+	if (this->rotation % 180 != 0) {
+		frame_width = primary->md.height;
+		frame_height = primary->md.width;
+	}
+	if (this->frame_width != frame_width ||
+	    this->frame_height != frame_height) {
+		this->frame_width = frame_width;
+		this->frame_height = frame_height;
+	}
+
+	g_signal_emit(this, sigs[SIG_FRAME], 0, length, bufs);
+
+out:
+	for (size_t i = 0; i < length; ++i)
+		for (int j = 0; j < bufs[i].md.length; ++j)
+			close(bufs[i].fds[j]);
+
+	if (ret <= 0) {
+		g_warning("Frame: Failed to receive frame message: %ld.", ret);
+		if (ret == 0)
+			g_warning("ReFrame Streamer disconnected.");
+	} else {
+		_schedule_frame_msg(this);
+	}
+
+	return ret;
+}
+
+static ssize_t _on_card_path_msg(RfStreamer *this)
+{
+	g_debug("DRM: Received card path message.");
+
+	size_t length = 0;
+	g_autofree char *card_path = NULL;
+	ssize_t ret = 0;
+	GInputStream *is =
+		g_io_stream_get_input_stream(G_IO_STREAM(this->connection));
+
+	ret = g_input_stream_read(is, &length, sizeof(length), NULL, NULL);
+	if (ret <= 0)
+		goto out;
+
+	card_path = g_malloc0(length);
+	ret = g_input_stream_read(is, card_path, length, NULL, NULL);
+	if (ret <= 0)
+		goto out;
+
+	g_signal_emit(this, sigs[SIG_CARD_PATH], 0, card_path);
+
+out:
+	if (ret <= 0)
+		g_warning("DRM: Failed to receive card path: %ld.", ret);
+	else
+		g_debug("DRM: Received card path: %s.", card_path);
+	return ret;
+}
+
 static gboolean
 _on_socket_in(GSocket *socket, GIOCondition condition, gpointer data)
 {
@@ -144,105 +253,34 @@ _on_socket_in(GSocket *socket, GIOCondition condition, gpointer data)
 	if (!(condition & this->io_flags))
 		return G_SOURCE_CONTINUE;
 
-	RfBuffer b;
-	GInputVector iov = { &b.md, sizeof(b.md) };
-	GSocketControlMessage **msgs = NULL;
-	int n_msgs = 0;
-	ssize_t ret;
-
-	ret = g_socket_receive_message(
-		socket, NULL, &iov, 1, &msgs, &n_msgs, NULL, NULL, NULL
-	);
-	if (ret == 0) {
-		g_warning("ReFrame Streamer disconnected.");
-		goto stop;
-	} else if (ret < 0) {
-		g_warning("Frame: Failed to receive frame from socket.");
-		goto stop;
+	ssize_t ret = 0;
+	GInputStream *is =
+		g_io_stream_get_input_stream(G_IO_STREAM(this->connection));
+	char type;
+	ret = g_input_stream_read(is, &type, sizeof(type), NULL, NULL);
+	if (ret <= 0) {
+		if (ret == 0)
+			g_warning("ReFrame Streamer disconnected.");
+		goto out;
 	}
 
-	for (int i = 0; i < RF_MAX_PLANES; ++i)
-		b.fds[i] = -1;
-
-	// We should only receive 1 message each time, but if we get many
-	// messages, always use the latest one to reduce lag.
-	for (int i = n_msgs - 1; i >= 0; --i) {
-		if (G_IS_UNIX_FD_MESSAGE(msgs[i])) {
-			GUnixFDMessage *msg = G_UNIX_FD_MESSAGE(msgs[i]);
-			GUnixFDList *fds = g_unix_fd_message_get_fd_list(msg);
-			for (int j = 0; j < b.md.length; ++j) {
-				b.fds[j] = g_unix_fd_list_get(fds, j, NULL);
-				// Some error happens.
-				if (b.fds[j] == -1) {
-					g_warning(
-						"Frame: Failed to receive frame fds from socket, only got %d of %d.",
-						j,
-						b.md.length
-					);
-					b.md.length = j;
-					goto cleanup;
-				}
-			}
-			// We don't need to free GUnixFDList because
-			// GUnixFDMessage does not return a reference so we
-			// are not taking ownership of it.
-			//
-			// See <https://docs.gtk.org/gio-unix/type_func.FDMessage.get_fd_list.html>.
-			break;
-		}
+	switch (type) {
+	case RF_MSG_TYPE_FRAME:
+		ret = _on_frame_msg(this);
+		break;
+	case RF_MSG_TYPE_CARD_PATH:
+		ret = _on_card_path_msg(this);
+		break;
+	default:
+		break;
 	}
 
-	g_debug("Frame: Got frame metadata: length %u, width %u, height %u, fourcc %c%c%c%c, modifier %#lx.",
-		b.md.length,
-		b.md.width,
-		b.md.height,
-		(b.md.fourcc >> 0) & 0xff,
-		(b.md.fourcc >> 8) & 0xff,
-		(b.md.fourcc >> 16) & 0xff,
-		(b.md.fourcc >> 24) & 0xff,
-		b.md.modifier);
-	g_debug("Frame: Got frame fds: %d %d %d %d.",
-		b.fds[0],
-		b.fds[1],
-		b.fds[2],
-		b.fds[3]);
-	g_debug("Frame: Got frame offsets: %u %u %u %u.",
-		b.md.offsets[0],
-		b.md.offsets[1],
-		b.md.offsets[2],
-		b.md.offsets[3]);
-	g_debug("Frame: Got frame pitches: %u %u %u %u.",
-		b.md.pitches[0],
-		b.md.pitches[1],
-		b.md.pitches[2],
-		b.md.pitches[3]);
-
-	if (this->width != b.md.width || this->height != b.md.height) {
-		if (this->rotation % 180 == 0) {
-			this->width = b.md.width;
-			this->height = b.md.height;
-		} else {
-			this->width = b.md.height;
-			this->height = b.md.width;
-		}
+out:
+	if (ret <= 0) {
+		rf_streamer_stop(this);
+		return G_SOURCE_REMOVE;
 	}
-
-	g_signal_emit(this, sigs[SIG_FRAME], 0, &b);
-
-cleanup:
-	for (int i = 0; i < b.md.length; ++i)
-		close(b.fds[i]);
-
-	for (int i = 0; i < n_msgs; ++i)
-		g_object_unref(msgs[i]);
-	g_free(msgs);
-
-	_schedule_frame_request(this);
 	return G_SOURCE_CONTINUE;
-
-stop:
-	rf_streamer_stop(this);
-	return G_SOURCE_REMOVE;
 }
 
 static void _dispose(GObject *o)
@@ -272,8 +310,8 @@ static void rf_streamer_init(RfStreamer *this)
 	this->monitor_x = 0;
 	this->monitor_y = 0;
 	this->rotation = 0;
-	this->width = 0;
-	this->height = 0;
+	this->frame_width = 0;
+	this->frame_height = 0;
 	this->running = false;
 }
 
@@ -298,8 +336,21 @@ static void rf_streamer_class_init(RfStreamerClass *klass)
 		NULL,
 		NULL,
 		G_TYPE_NONE,
+		2,
+		G_TYPE_INT,
+		G_TYPE_POINTER
+	);
+	sigs[SIG_CARD_PATH] = g_signal_new(
+		"card-path",
+		RF_TYPE_STREAMER,
+		0,
+		0,
+		NULL,
+		NULL,
+		NULL,
+		G_TYPE_NONE,
 		1,
-		RF_TYPE_BUFFER
+		G_TYPE_STRING
 	);
 }
 
@@ -348,8 +399,8 @@ int rf_streamer_start(RfStreamer *this)
 	);
 	this->rotation = rf_config_get_rotation(this->config);
 	g_message("Frame: Got screen rotation %u.", this->rotation);
-	this->width = 0;
-	this->height = 0;
+	this->frame_width = 0;
+	this->frame_height = 0;
 	this->connection = g_socket_client_connect(
 		this->client, G_SOCKET_CONNECTABLE(this->address), NULL, NULL
 	);
@@ -363,7 +414,7 @@ int rf_streamer_start(RfStreamer *this)
 		this->source, G_SOURCE_FUNC(_on_socket_in), this, NULL
 	);
 	g_source_attach(this->source, NULL);
-	_schedule_frame_request(this);
+	_schedule_frame_msg(this);
 
 	this->running = true;
 	g_debug("Emitting ReFrame Streamer start signal.");
@@ -413,9 +464,8 @@ void rf_streamer_send_keyboard_event(
 	if (!this->running)
 		return;
 
-#define KEYBOARD_EVENT_LENGTH 2
-	struct input_event ies[KEYBOARD_EVENT_LENGTH];
-	memset(ies, 0, KEYBOARD_EVENT_LENGTH * sizeof(*ies));
+	struct input_event ies[KEYBOARD_MAX_EVENTS];
+	memset(ies, 0, KEYBOARD_MAX_EVENTS * sizeof(*ies));
 
 	ies[0].type = EV_KEY;
 	ies[0].code = keycode;
@@ -425,7 +475,7 @@ void rf_streamer_send_keyboard_event(
 	ies[1].code = SYN_REPORT;
 	ies[1].value = 0;
 
-	_send_input_request(this, ies, KEYBOARD_EVENT_LENGTH);
+	_send_input_msg(this, ies, KEYBOARD_MAX_EVENTS);
 }
 
 void rf_streamer_send_pointer_event(
@@ -447,19 +497,23 @@ void rf_streamer_send_pointer_event(
 	// Assuming user only have 1 monitor when they set desktop size to 0x0.
 	const int desktop_width = this->desktop_width > 0 ?
 					  this->desktop_width :
-					  this->width + this->monitor_x;
+					  this->monitor_x + this->frame_width;
 	const int desktop_height = this->desktop_height > 0 ?
 					   this->desktop_height :
-					   this->height + this->monitor_y;
+					   this->monitor_y + this->frame_height;
 	// Typically desktop environment will map uinput `EV_ABS` max size to
 	// the whole virtual desktop, so we need to convert the position to
 	// global position in the virtual desktop.
-	const double x = (rx * this->width + this->monitor_x) / desktop_width;
-	const double y = (ry * this->height + this->monitor_y) / desktop_height;
+	const double x =
+		(this->monitor_x + rx * this->frame_width) / desktop_width;
+	const double y =
+		(this->monitor_y + ry * this->frame_height) / desktop_height;
 	g_debug("Input: Calculated global position x %f and y %f.", x, y);
 
-	size_t length = (wup || wdown) ? 7 : 6;
-	g_autofree struct input_event *ies = g_malloc0_n(length, sizeof(*ies));
+	size_t length = (wup || wdown) ? POINTER_MAX_EVENTS :
+					 POINTER_MAX_EVENTS - 1;
+	struct input_event ies[POINTER_MAX_EVENTS];
+	memset(ies, 0, POINTER_MAX_EVENTS * sizeof(*ies));
 
 	ies[0].type = EV_ABS;
 	ies[0].code = ABS_X;
@@ -491,5 +545,5 @@ void rf_streamer_send_pointer_event(
 	ies[length - 1].code = SYN_REPORT;
 	ies[length - 1].value = 0;
 
-	_send_input_request(this, ies, length);
+	_send_input_msg(this, ies, length);
 }
