@@ -55,80 +55,95 @@
 struct _this {
 	RfConfig *config;
 	GSocketConnection *connection;
+	char *card_path;
 	int cfd;
-	uint32_t connector_id;
+	uint32_t crtc_id;
+	uint32_t primary_id;
+	bool cursor;
+	uint32_t cursor_id;
 	int ufd;
 };
 
-static uint32_t
-_get_connector_id(int cfd, drmModeRes *res, const char *connector_name)
+// You need to explicitly cast the type of returned value.
+static uint64_t _get_plane_prop(
+	int cfd,
+	uint32_t plane_id,
+	const char *name,
+	uint64_t default_value
+)
 {
-	uint32_t id = 0;
-	bool found = false;
-	for (int i = 0; i < res->count_connectors; ++i) {
-		drmModeConnector *connector =
-			drmModeGetConnector(cfd, res->connectors[i]);
-		g_autofree char *full_name = g_strdup_printf(
-			"%s-%d",
-			drmModeGetConnectorTypeName(connector->connector_type),
-			connector->connector_type_id
-		);
-		g_debug("DRM: Connector %s is %s.",
-			full_name,
-			connector->connection == DRM_MODE_CONNECTED ?
-				"connected" :
-				"disconnected");
-		id = connector->connector_id;
-		drmModeFreeConnector(connector);
-		if (connector->connection == DRM_MODE_CONNECTED &&
-		    g_strcmp0(full_name, connector_name) == 0) {
-			found = true;
-			break;
-		}
+	uint64_t value = default_value;
+
+	drmModeObjectProperties *props = drmModeObjectGetProperties(
+		cfd, plane_id, DRM_MODE_OBJECT_PLANE
+	);
+	if (props == NULL)
+		return -1;
+
+	for (size_t i = 0; i < props->count_props; ++i) {
+		drmModePropertyRes *prop =
+			drmModeGetProperty(cfd, props->props[i]);
+		if (prop == NULL)
+			continue;
+		if (g_strcmp0(prop->name, name) == 0)
+			value = props->prop_values[i];
+		drmModeFreeProperty(prop);
 	}
-	return found ? id : 0;
+
+	drmModeFreeObjectProperties(props);
+
+	return value;
 }
 
-static uint32_t _get_fb_id(int cfd, uint32_t connector_id)
+// We have to get plane via ID every frame to get the newest framebuffer.
+static uint32_t _get_plane_id(int cfd, uint32_t crtc_id, uint32_t type)
 {
-	uint32_t id = 0;
-	drmModeConnector *connector = NULL;
-	drmModeEncoder *encoder = NULL;
-	drmModeCrtc *crtc = NULL;
+	uint32_t plane_id = 0;
 
-	connector = drmModeGetConnector(cfd, connector_id);
-	if (connector == NULL)
-		goto clean_connector;
-	encoder = drmModeGetEncoder(cfd, connector->encoder_id);
-	if (encoder == NULL)
-		goto clean_encoder;
-	crtc = drmModeGetCrtc(cfd, encoder->crtc_id);
-	if (crtc == NULL)
-		goto clean_crtc;
-	id = crtc->buffer_id;
+	drmModePlaneRes *pres = drmModeGetPlaneResources(cfd);
+	if (pres == NULL) {
+		g_warning("DRM: Failed to get plane resources.");
+		return 0;
+	}
 
-clean_crtc:
-	drmModeFreeCrtc(crtc);
-clean_encoder:
-	drmModeFreeEncoder(encoder);
-clean_connector:
-	drmModeFreeConnector(connector);
+	g_debug("DRM: Finding plane of type %s and CRTC ID %u.",
+		rf_plane_type(type),
+		crtc_id);
 
-	return id;
+	for (size_t i = 0; i < pres->count_planes; ++i) {
+		if (_get_plane_prop(
+			    cfd, pres->planes[i], "type", DRM_PLANE_TYPE_OVERLAY
+		    ) != type)
+			continue;
+		drmModePlane *plane = drmModeGetPlane(cfd, pres->planes[i]);
+		if (plane == NULL)
+			continue;
+		// Ignore unrelated planes.
+		g_debug("DRM: Plane ID %u is of type %s belongs to CRTC ID %u.",
+			plane->plane_id,
+			rf_plane_type(type),
+			plane->crtc_id);
+		if (plane->crtc_id == crtc_id)
+			plane_id = plane->plane_id;
+		drmModeFreePlane(plane);
+		if (plane_id != 0)
+			break;
+	}
+
+	drmModeFreePlaneResources(pres);
+	return plane_id;
 }
 
-static int _export_fb2(struct _this *this, uint32_t fb_id, RfBuffer *b)
+static int _export_fb2(int cfd, RfBuffer *b, uint32_t fb_id)
 {
-	drmModeFB2 *fb = drmModeGetFB2(this->cfd, fb_id);
+	drmModeFB2 *fb = drmModeGetFB2(cfd, fb_id);
 	if (fb == NULL)
 		return -1;
-	g_debug("Frame: Got FB2 frame %u.", fb->fb_id);
-	for (int i = 0; i < RF_MAX_PLANES; ++i) {
+	g_debug("Frame: Got FB2 framebuffer ID %u.", fb->fb_id);
+	for (int i = 0; i < RF_MAX_FDS; ++i) {
 		if (fb->handles[i] == 0)
 			break;
-		drmPrimeHandleToFD(
-			this->cfd, fb->handles[i], DRM_CLOEXEC, &b->fds[i]
-		);
+		drmPrimeHandleToFD(cfd, fb->handles[i], DRM_CLOEXEC, &b->fds[i]);
 		if (b->fds[i] < 0)
 			break;
 		++b->md.length;
@@ -147,15 +162,15 @@ static int _export_fb2(struct _this *this, uint32_t fb_id, RfBuffer *b)
 	return b->md.length;
 }
 
-static int _export_fb(struct _this *this, uint32_t fb_id, RfBuffer *b)
+static int _export_fb(int cfd, RfBuffer *b, uint32_t fb_id)
 {
-	drmModeFB *fb = drmModeGetFB(this->cfd, fb_id);
+	drmModeFB *fb = drmModeGetFB(cfd, fb_id);
 	if (fb == NULL)
 		return -1;
-	g_debug("Frame: Got FB frame %u.", fb->fb_id);
+	g_debug("Frame: Got FB framebuffer ID %u.", fb->fb_id);
 	if (fb->handle == 0)
 		return 0;
-	drmPrimeHandleToFD(this->cfd, fb->handle, DRM_CLOEXEC, &b->fds[0]);
+	drmPrimeHandleToFD(cfd, fb->handle, DRM_CLOEXEC, &b->fds[0]);
 	if (b->fds[0] < 0)
 		return 0;
 	b->md.length = 1;
@@ -169,82 +184,137 @@ static int _export_fb(struct _this *this, uint32_t fb_id, RfBuffer *b)
 	return b->md.length;
 }
 
-static ssize_t _on_frame_request(struct _this *this)
+static int _make_buffer(int cfd, RfBuffer *b, uint32_t plane_id, uint32_t type)
 {
-	g_debug("Frame: Received frame request.");
+	drmModePlane *plane = drmModeGetPlane(cfd, plane_id);
+	if (plane == NULL)
+		return 0;
+	uint32_t fb_id = plane->fb_id;
+	b->md.type = type;
+	b->md.crtc_x = (int64_t)_get_plane_prop(cfd, plane_id, "CRTC_X", 0);
+	b->md.crtc_y = (int64_t)_get_plane_prop(cfd, plane_id, "CRTC_Y", 0);
+	drmModeFreePlane(plane);
+	if (fb_id == 0)
+		return 0;
+	g_debug("Frame: Got %s plane framebuffer ID %u.",
+		rf_plane_type(type),
+		fb_id);
 
-	uint32_t fb_id = _get_fb_id(this->cfd, this->connector_id);
-	ssize_t ret = 0;
-	RfBuffer b;
+	int ret = 0;
 	// GUnixFDList refuses to send invalid fds like -1, so we need
 	// to pass the number of fds. We assume valid planes are
 	// continuous.
-	for (int i = 0; i < RF_MAX_PLANES; ++i) {
-		b.fds[i] = -1;
-		b.md.offsets[i] = 0;
-		b.md.pitches[i] = 0;
+	b->md.length = 0;
+	for (int i = 0; i < RF_MAX_FDS; ++i) {
+		b->fds[i] = -1;
+		b->md.offsets[i] = 0;
+		b->md.pitches[i] = 0;
 	}
-	b.md.length = 0;
 	// Export DRM framebuffer to fds and metadata that EGL can import.
-	ret = _export_fb2(this, fb_id, &b);
+	ret = _export_fb2(cfd, b, fb_id);
 	if (ret <= 0)
-		ret = _export_fb(this, fb_id, &b);
+		ret = _export_fb(cfd, b, fb_id);
 	if (ret <= 0) {
 		g_warning("Frame: Failed to get frame.");
 		return ret;
 	}
+	rf_buffer_debug(b);
+	return ret;
+}
 
-	g_debug("Frame: Got frame metadata: length %u, width %u, height %u, fourcc %c%c%c%c, modifier %#lx.",
-		b.md.length,
-		b.md.width,
-		b.md.height,
-		(b.md.fourcc >> 0) & 0xff,
-		(b.md.fourcc >> 8) & 0xff,
-		(b.md.fourcc >> 16) & 0xff,
-		(b.md.fourcc >> 24) & 0xff,
-		b.md.modifier);
-	g_debug("Frame: Got frame fds: %d %d %d %d.",
-		b.fds[0],
-		b.fds[1],
-		b.fds[2],
-		b.fds[3]);
-	g_debug("Frame: Got frame offsets: %u %u %u %u.",
-		b.md.offsets[0],
-		b.md.offsets[1],
-		b.md.offsets[2],
-		b.md.offsets[3]);
-	g_debug("Frame: Got frame pitches: %u %u %u %u.",
-		b.md.pitches[0],
-		b.md.pitches[1],
-		b.md.pitches[2],
-		b.md.pitches[3]);
-
-	GOutputVector iov = { &b.md, sizeof(b.md) };
+static ssize_t _send_buffer(GSocketConnection *connection, RfBuffer *b)
+{
+	ssize_t ret = 0;
+	GOutputVector iov = { &b->md, sizeof(b->md) };
 	GUnixFDList *fds = g_unix_fd_list_new();
 	// This won't take the ownership so we need to close fds.
 	//
 	// See <https://docs.gtk.org/gio/method.UnixFDList.append.html>.
-	for (int i = 0; i < b.md.length; ++i)
-		g_unix_fd_list_append(fds, b.fds[i], NULL);
+	for (int i = 0; i < b->md.length; ++i)
+		g_unix_fd_list_append(fds, b->fds[i], NULL);
 	// This won't take the ownership so we need to free GUnixFDList.
 	GSocketControlMessage *msg = g_unix_fd_message_new_with_fd_list(fds);
-	GSocket *socket = g_socket_connection_get_socket(this->connection);
+	GSocket *socket = g_socket_connection_get_socket(connection);
 	ret = g_socket_send_message(
 		socket, NULL, &iov, 1, &msg, 1, G_SOCKET_MSG_NONE, NULL, NULL
 	);
 	g_object_unref(fds);
 	g_object_unref(msg);
-
-	for (int i = 0; i < b.md.length; ++i)
-		close(b.fds[i]);
-
-	if (ret < 0)
-		g_warning("Frame: Failed to send frame to socket.");
 	return ret;
 }
 
-static ssize_t _on_input_request(struct _this *this)
+static ssize_t
+_send_frame_msg(struct _this *this, size_t length, RfBuffer *bufs)
 {
+	ssize_t ret = 0;
+	ret = rf_send_header(this->connection, RF_MSG_TYPE_FRAME, length);
+	if (ret <= 0)
+		return ret;
+
+	for (size_t i = 0; i < length; ++i) {
+		ret = _send_buffer(this->connection, &bufs[i]);
+		if (ret <= 0)
+			break;
+	}
+	return ret;
+}
+
+static ssize_t _on_frame_msg(struct _this *this)
+{
+	g_debug("Frame: Received frame message.");
+
+	size_t length = 0;
+	ssize_t ret = 0;
+	GInputStream *is =
+		g_io_stream_get_input_stream(G_IO_STREAM(this->connection));
+
+	ret = g_input_stream_read(is, &length, sizeof(length), NULL, NULL);
+	if (ret <= 0)
+		goto out;
+
+	RfBuffer bufs[RF_MAX_BUFS];
+	length = 0;
+	ret = _make_buffer(
+		this->cfd,
+		&bufs[length++],
+		this->primary_id,
+		DRM_PLANE_TYPE_PRIMARY
+	);
+	if (ret <= 0)
+		return ret;
+
+	if (this->cursor && this->cursor_id == 0)
+		this->cursor_id = _get_plane_id(
+			this->cfd, this->crtc_id, DRM_PLANE_TYPE_CURSOR
+		);
+	if (this->cursor_id != 0) {
+		ret = _make_buffer(
+			this->cfd,
+			&bufs[length++],
+			this->cursor_id,
+			DRM_PLANE_TYPE_CURSOR
+		);
+		// It is OK to ignore cursor plane if failed.
+		if (ret <= 0)
+			--length;
+	}
+
+	ret = _send_frame_msg(this, length, bufs);
+
+	for (size_t i = 0; i < length; ++i)
+		for (int j = 0; j < bufs[i].md.length; ++j)
+			close(bufs[i].fds[j]);
+
+out:
+	if (ret <= 0)
+		g_warning("Frame: Failed to send/receive frame: %ld.", ret);
+	return ret;
+}
+
+static ssize_t _on_input_msg(struct _this *this)
+{
+	g_debug("Input: Received input message.");
+
 	g_autofree struct input_event *ies = NULL;
 	size_t length = 0;
 	ssize_t ret = 0;
@@ -252,57 +322,211 @@ static ssize_t _on_input_request(struct _this *this)
 		g_io_stream_get_input_stream(G_IO_STREAM(this->connection));
 
 	ret = g_input_stream_read(is, &length, sizeof(length), NULL, NULL);
-	if (ret <= 0) {
-		if (ret < 0)
-			g_warning(
-				"Input: Failed to receive input events length from socket."
-			);
-		return ret;
-	}
+	if (ret <= 0)
+		goto out;
 	ies = g_malloc_n(length, sizeof(*ies));
 	ret = g_input_stream_read(is, ies, length * sizeof(*ies), NULL, NULL);
-	if (ret < 0) {
-		if (ret < 0)
-			g_warning(
-				"Input: Failed to receive %lu * %ld bytes input events length from socket.",
-				length,
-				sizeof(*ies)
-			);
-		return ret;
-	}
-	g_debug("Input: Received %lu * %ld bytes input events.",
-		length,
-		sizeof(*ies));
+	if (ret <= 0)
+		goto out;
+
 	_write_may(this->ufd, ies, length * sizeof(*ies));
 
+out:
+	if (ret <= 0)
+		g_warning("Input: Failed to receive input events: %ld.", ret);
+	else
+		g_debug("Input: Received %lu * %ld bytes input events.",
+			length,
+			sizeof(*ies));
 	return ret;
 }
 
-static void _init_drm(struct _this *this)
+static ssize_t _send_card_path_msg(struct _this *this, const char *card_path)
 {
-	g_autofree char *card_path = rf_config_get_card_path(this->config);
-	this->cfd = open(card_path, O_RDONLY | O_CLOEXEC);
-	if (this->cfd <= 0)
-		g_error("DRM: Failed to open card %s: %s.",
-			card_path,
-			strerror(errno));
-	g_message("DRM: Opened card %s.", card_path);
-
-	drmModeRes *res = NULL;
-	g_autofree char *connector_name = rf_config_get_connector(this->config);
-	res = drmModeGetResources(this->cfd);
-	this->connector_id = _get_connector_id(this->cfd, res, connector_name);
-	drmModeFreeResources(res);
-	if (this->connector_id == 0)
-		g_error("DRM: Failed to find a connected connector %s.",
-			connector_name);
-	g_message("DRM: Found connected connector %s.", connector_name);
+	// Send the `\0` so we could easily use it in receiver.
+	size_t length = strlen(card_path) + 1;
+	ssize_t ret = 0;
+	GOutputStream *os =
+		g_io_stream_get_output_stream(G_IO_STREAM(this->connection));
+	ret = rf_send_header(this->connection, RF_MSG_TYPE_CARD_PATH, length);
+	if (ret <= 0)
+		return ret;
+	ret = g_output_stream_write(os, card_path, length, NULL, NULL);
+	return ret;
 }
 
-static void _init_uinput(struct _this *this)
+static inline char *_get_connector_name(drmModeConnector *connector)
+{
+	return g_strdup_printf(
+		"%s-%d",
+		drmModeGetConnectorTypeName(connector->connector_type),
+		connector->connector_type_id
+	);
+}
+
+static drmModeConnector *_get_connector(int cfd, const char *connector_name)
+{
+	drmModeConnector *connector = NULL;
+
+	drmModeRes *res = drmModeGetResources(cfd);
+	if (res == NULL) {
+		g_warning("DRM: Failed to get resources.");
+		return NULL;
+	}
+
+	if (connector_name != NULL)
+		g_debug("DRM: Finding connector for connector %s.",
+			connector_name);
+
+	for (int i = 0; i < res->count_connectors; ++i) {
+		connector = drmModeGetConnector(cfd, res->connectors[i]);
+		if (connector == NULL)
+			continue;
+		g_autofree char *full_name = _get_connector_name(connector);
+		g_debug("DRM: Connector %s is %s.",
+			full_name,
+			connector->connection == DRM_MODE_CONNECTED ?
+				"connected" :
+				"disconnected");
+		if (connector->connection == DRM_MODE_CONNECTED &&
+		    (connector_name == NULL ||
+		     g_strcmp0(full_name, connector_name) == 0))
+			break;
+		drmModeFreeConnector(connector);
+	}
+
+	drmModeFreeResources(res);
+	return connector;
+}
+
+static drmModeConnector *
+_get_connected_card_and_connector(struct _this *this, const char *connector_name)
+{
+	g_autoptr(GDir) dir = g_dir_open("/dev/dri", 0, NULL);
+	if (dir == NULL)
+		return NULL;
+
+	const char *name = NULL;
+	while ((name = g_dir_read_name(dir)) != NULL) {
+		if (!g_str_has_prefix(name, "card"))
+			continue;
+		g_autofree char *card_path =
+			g_strdup_printf("/dev/dri/%s", name);
+		int cfd = open(card_path, O_RDONLY | O_CLOEXEC);
+		if (cfd < 0)
+			continue;
+		g_debug("DRM: Finding the first connected connector on card %s.",
+			card_path);
+		drmModeConnector *connector =
+			_get_connector(cfd, connector_name);
+		if (connector != NULL) {
+			this->cfd = cfd;
+			this->card_path = g_strdup(card_path);
+			g_message(
+				"DRM: Found the first connected connector on card %s.",
+				card_path
+			);
+			return connector;
+		}
+		close(cfd);
+	}
+	return NULL;
+}
+
+static drmModeConnector *
+_get_card_and_connector(struct _this *this, const char *connector_name)
+{
+	if (this->card_path == NULL)
+		return _get_connected_card_and_connector(this, connector_name);
+
+	this->cfd = open(this->card_path, O_RDONLY | O_CLOEXEC);
+	if (this->cfd < 0) {
+		g_warning(
+			"DRM: Failed to open card %s: %s.",
+			this->card_path,
+			strerror(errno)
+		);
+		return NULL;
+	}
+	g_message("DRM: Opened card %s.", this->card_path);
+
+	return _get_connector(this->cfd, connector_name);
+}
+
+static drmModeCrtc *_get_crtc(int cfd, drmModeConnector *connector)
+{
+	drmModeEncoder *encoder = NULL;
+	drmModeCrtc *crtc = NULL;
+
+	encoder = drmModeGetEncoder(cfd, connector->encoder_id);
+	if (encoder == NULL)
+		return NULL;
+
+	crtc = drmModeGetCrtc(cfd, encoder->crtc_id);
+
+	drmModeFreeEncoder(encoder);
+
+	return crtc;
+}
+
+static void _setup_drm(struct _this *this)
+{
+	this->primary_id = 0;
+	this->cursor_id = 0;
+	this->cursor = true;
+
+	this->crtc_id = 0;
+
+	this->card_path = rf_config_get_card_path(this->config);
+	g_autofree char *connector_name = rf_config_get_connector(this->config);
+	drmModeConnector *connector =
+		_get_card_and_connector(this, connector_name);
+	if (connector == NULL)
+		g_error("DRM: Failed to find a connected connector.");
+	if (connector_name == NULL)
+		connector_name = _get_connector_name(connector);
+	g_message("DRM: Found connected connector %s.", connector_name);
+
+	drmModeCrtc *crtc = _get_crtc(this->cfd, connector);
+	drmModeFreeConnector(connector);
+	if (crtc == NULL)
+		g_error("DRM: Failed to find a CRTC for connector.");
+	this->crtc_id = crtc->crtc_id;
+	drmModeFreeCrtc(crtc);
+
+	// This is needed to get primary and cursor planes.
+	if (drmSetClientCap(this->cfd, DRM_CLIENT_CAP_UNIVERSAL_PLANES, 1) < 0)
+		g_warning("DRM: Failed to set universal planes capability.");
+
+	// This is needed to get `CRTC_X/Y` properties of planes.
+	if (drmSetClientCap(this->cfd, DRM_CLIENT_CAP_ATOMIC, 1) < 0)
+		g_warning("DRM: Failed to set atomic capability.");
+
+	this->primary_id =
+		_get_plane_id(this->cfd, this->crtc_id, DRM_PLANE_TYPE_PRIMARY);
+	if (this->primary_id == 0)
+		g_error("DRM: Failed to find a primary plane for CRTC.");
+
+	this->cursor = rf_config_get_cursor(this->config);
+	g_message(
+		"DRM: Cursor plane is %s.",
+		this->cursor ? "enabled" : "disabled"
+	);
+}
+
+static void _clean_drm(struct _this *this)
+{
+	if (this->cfd >= 0) {
+		close(this->cfd);
+		this->cfd = -1;
+	}
+	g_clear_pointer(&this->card_path, g_free);
+}
+
+static void _setup_uinput(struct _this *this)
 {
 	this->ufd = open("/dev/uinput", O_WRONLY | O_NONBLOCK);
-	if (this->ufd <= 0)
+	if (this->ufd < 0)
 		g_error("Input: Failed to open uinput: %s.", strerror(errno));
 
 	_ioctl_must(this->ufd, UI_SET_EVBIT, EV_KEY);
@@ -342,6 +566,15 @@ static void _init_uinput(struct _this *this)
 	_ioctl_must(this->ufd, UI_DEV_CREATE);
 }
 
+static void _clean_uinput(struct _this *this)
+{
+	if (this->ufd >= 0) {
+		_ioctl_may(this->ufd, UI_DEV_DESTROY);
+		close(this->ufd);
+		this->ufd = -1;
+	}
+}
+
 int main(int argc, char *argv[])
 {
 	g_autofree char *config_path = NULL;
@@ -408,6 +641,8 @@ int main(int argc, char *argv[])
 	g_autofree struct _this *this = g_malloc0(sizeof(*this));
 	g_message("Using configuration file %s.", config_path);
 	this->config = rf_config_new(config_path);
+	this->cfd = -1;
+	this->ufd = -1;
 
 	g_autoptr(GSocketListener) listener = g_socket_listener_new();
 	g_message("Using socket %s.", socket_path);
@@ -456,53 +691,41 @@ int main(int argc, char *argv[])
 
 		g_message("ReFrame Server connected.");
 
-		_init_drm(this);
-		_init_uinput(this);
+		_setup_drm(this);
+		_setup_uinput(this);
+
+		_send_card_path_msg(this, this->card_path);
 
 		while (true) {
 			ssize_t ret = 0;
 			GInputStream *is = g_io_stream_get_input_stream(
 				G_IO_STREAM(this->connection)
 			);
-			char request;
+			char type;
 			ret = g_input_stream_read(
-				is, &request, sizeof(request), NULL, NULL
+				is, &type, sizeof(type), NULL, NULL
 			);
-			if (ret == 0) {
-				g_message("ReFrame Server disconnected.");
+			if (ret <= 0)
 				break;
-			} else if (ret < 0) {
-				g_warning(
-					"Failed to receive request from socket."
-				);
-				break;
-			}
 
-			switch (request) {
-			case RF_REQUEST_TYPE_FRAME:
-				ret = _on_frame_request(this);
+			switch (type) {
+			case RF_MSG_TYPE_FRAME:
+				ret = _on_frame_msg(this);
 				break;
-			case RF_REQUEST_TYPE_INPUT:
-				ret = _on_input_request(this);
+			case RF_MSG_TYPE_INPUT:
+				ret = _on_input_msg(this);
 				break;
 			default:
 				break;
 			}
-			if (ret <= 0) {
-				if (ret == 0)
-					g_message(
-						"ReFrame Server disconnected."
-					);
+			if (ret <= 0)
 				break;
-			}
 		}
 
-		_ioctl_may(this->ufd, UI_DEV_DESTROY);
-		close(this->ufd);
-		this->ufd = 0;
+		g_message("ReFrame Server disconnected.");
 
-		close(this->cfd);
-		this->cfd = 0;
+		_clean_uinput(this);
+		_clean_drm(this);
 
 		g_io_stream_close(G_IO_STREAM(this->connection), NULL, &error);
 		if (error != NULL)
